/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CMySQLDatabase.h  -  description
                             -------------------
    begin                : Mon Sep 17 2007
    copyright            : (C) 2008 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors 
    may be used to endorse or promote products derived from this software 
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CMYSQLDATABASE_H__
#define __CMYSQLDATABASE_H__

#include <sptk3/db/CDatabase.h>

#if HAVE_MYSQL == 1

#include <mysql/mysql.h>

namespace sptk {

    /// @addtogroup Database Database Support
    /// @{

    class CQuery;
    class CMySQLStatement;

    /// @brief MySQL database
    ///
    /// CMySQLDatabase is thread-safe connection to MySQL database.
    class CMySQLDatabase : public CDatabase {
        friend class CQuery;

    private:

        MYSQL*       m_connect;  ///< MySQL database connection

        std::string  m_host;
        std::string  m_user;
        std::string  m_password;
        std::string  m_dbName;
        uint32_t     m_port;

        /// @brief Parses connection string (a set of space-separated pairs name=value)
        /// @param connectionString std::string, a connection string
        /// @param host std::string&, (output) host name
        /// @param user std::string&, (output) user name
        /// @param password std::string&, (output) user password
        /// @param dbName std::string&, (output) database name
        /// @param port uint32_t&, (output) MySQL server port number
        void parseConnectionString(std::string connectionString,
            std::string& host, std::string& user, std::string& password,
            std::string& dbName, uint32_t& port) const;

    protected:

        /// @brief Begins the transaction
        virtual void driverBeginTransaction() throw(CException);

        /// @brief Ends the transaction
        /// @param commit bool, commit if true, rollback if false
        virtual void driverEndTransaction(bool commit) throw(CException);

        // These methods implement the actions requested by CQuery
        virtual std::string queryError(const CQuery *query) const; ///< Retrieves an error (if any) after executing a statement
        virtual void queryAllocStmt(CQuery *query);      ///< Allocates an MySQL statement
        virtual void queryFreeStmt(CQuery *query);       ///< Deallocates an MySQL statement
        virtual void queryCloseStmt(CQuery *query);      ///< Closes an MySQL statement
        virtual void queryPrepare(CQuery *query);        ///< Prepares a query if supported by database
        virtual void queryUnprepare(CQuery *query);      ///< Unprepares a query if supported by database
        virtual void queryExecute(CQuery *query);        ///< Executes a statement
        virtual int  queryColCount(CQuery *query);       ///< Counts columns of the dataset (if any) returned by query
        virtual void queryBindParameters(CQuery *query); ///< Binds the parameters to the query
        virtual void queryOpen(CQuery *query);           ///< Opens the query for reading data from the query' recordset
        virtual void queryFetch(CQuery *query);          ///< Reads data from the query' recordset into fields, and advances to the next row. After reading the last row sets the EOF (end of file, or no more data) flag.

        /// @brief Returns parameter mark
        ///
        /// Parameter mark is generated from the parameterIndex.
        /// @param paramIndex unsigned, parameter index in SQL starting from 0
        virtual std::string paramMark(unsigned paramIndex);

    public:
        /// @brief Returns the MySQL connection object
        MYSQL* connection() {
            return m_connect;
        }

    public:
        /// @brief Returns true if the MySQL type is a buffer
        static bool isMySQLTypeBuffer(enum_field_types mysqlType);

        /// @brief Converts datatype from MySQL type to SPTK CVariantType
        static void MySQLTypeToCType(enum_field_types mysqlType, CVariantType& dataType);

        /// @brief Converts datatype from SPTK CVariantType to MySQL type
        static void CTypeToMySQLType(CVariantType dataType, enum_field_types& mysqlType);

    public:

        /// @brief Constructor
        ///
        /// Typical connection string is something like: "dbname='mydb' host='myhostname' port=5142" and so on.
        /// For more information please refer to:
        /// http://www.postgresql.org/docs/current/interactive/libpq-connect.html
        /// If the connection string is empty then default database with the name equal to user name is used.
        /// @param connectionString std::string, the MySQL connection string
        CMySQLDatabase(std::string connectionString = "");

        /// @brief Destructor
        virtual ~CMySQLDatabase();

        /// @brief Opens the database connection. If unsuccessful throws an exception.
        /// @param connectionString std::string, the MySQL connection string
        virtual void openDatabase(std::string connectionString = "") throw(CException);

        /// @brief Closes the database connection. If unsuccessful throws an exception.
        virtual void closeDatabase() throw(CException);

        /// @brief Returns true if database is opened
        virtual bool active() const;

        /// @brief Returns the database connection handle
        virtual void* handle() const;

        /// @brief Returns the MySQL driver description for the active connection
        virtual std::string driverDescription() const;

        /// @brief Lists database objects
        /// @param objectType CDbObjectType, object type to list
        /// @param objects CStrings&, object list (output)
        virtual void objectList(CDbObjectType objectType, CStrings& objects) throw(std::exception);
    };

    /// @brief MySQL synchronization object
    ///
    /// Locks MySQL database. That is needed for thread safety.
    /// Defines only protected constructors to be used internally.
    class CMySQLLock {
        friend class CMySQLDatabase;

        CMySQLDatabase& m_object;    /// MySQL database
    protected:
        /// @brief Protected constructor
        ///
        /// Locks the MySQL object till the destructor is called.
        /// @param object CMySQLDatabase&, MySQL object to lock.
        CMySQLLock(CMySQLDatabase& object) : m_object(object) {
            if (&m_object)
                m_object.lock();
        }

        /// @brief Protected constructor
        ///
        /// Locks the MySQL object till the destructor is called.
        /// @param object CMySQLDatabase*, MySQL object to lock.
        CMySQLLock(CMySQLDatabase* object) : m_object(*object) {
            if (&m_object)
                m_object.lock();
        }
    public:

        /// @brief Destructor
        ///
        /// Unlocks the MySQL object defined in constructor.
        ~CMySQLLock() {
            if (&m_object)
                m_object.unlock();
        }
    };

/// @}
}

#endif

#endif
