/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CThemeColorCollection.h  -  description
                             -------------------
    begin                : Sat Jul 03 2008
    copyright            : (C) 2003-2008 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.
 * Neither the name of the <ORGANIZATION> nor the names of its contributors 
   may be used to endorse or promote products derived from this software 
   without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CTHEMECOLORCOLLECTION_H__
#define __CTHEMECOLORCOLLECTION_H__

#include <FL/Fl.h>
#include <FL/Enumerations.h>
#include <sptk3/cxml>
#include <sptk3/gui/CThemeImageState.h>
#include <map>
#include <string>

namespace sptk {

/// @addtogroup gui GUI Classes
/// @{

/// @brief Constants for color indices
enum CThemeColorIndex {
    THM_FOREGROUND_COLOR,   ///< Foreground color
    THM_BACKGROUND_COLOR,   ///< Background color
    THM_BASE_COLOR,         ///< Base color
    THM_TEXT_COLOR          ///< Text color
};

/// @brief Number of color indices available
#define THM_MAX_COLOR_INDEX 4

/// @brief Color variations
enum CThemeColorState {
    THM_COLOR_UNDEFINED=-1, ///< Undefined color
    THM_COLOR_NORMAL=0,     ///< Normal color
    THM_COLOR_PRELIGHT,     ///< Prelight color
    THM_COLOR_SELECTED,     ///< Selected element color
    THM_COLOR_ACTIVE,       ///< Active element color
    THM_COLOR_INSENSITIVE   ///< Insensitive (disabled) element color
};

/// @brief Number of color states available
#define THM_MAX_COLOR_STATE 5

/// @brief Type definition for color modification functions
typedef Fl_Color (*gtk_color_function)(std::string expression);

/// @brief Defines theme color map
class CThemeColorCollection {
    static std::map<std::string,gtk_color_function> m_gtkColorFunctionMap;  ///< The color modification functions map
    static std::map<std::string,Fl_Color> m_colorMap;                       ///< Color name to value map
    Fl_Color    m_colors[THM_MAX_COLOR_INDEX][MAX_IMAGE_STATES];            ///< Current color variations array
    static Fl_Color gtkColorFunction(std::string expression);               ///< Computes color from given expression
    void loadColor(CXmlNode* colorNode,CThemeColorIndex colorIndex);        ///< Loads a color from theme color node
    bool loadColorMap(CXmlDoc& gtkTheme,std::string colorMapXPath);         ///< Loads a color map from theme

    /// @brief Color modification function: no modification
    /// @param expression std::string, color expression
    static Fl_Color passby(std::string expression);

    /// @brief Color modification function: lighter color
    /// @param expression std::string, color expression
    static Fl_Color lighter(std::string expression);

    /// @brief Color modification function: darker color
    /// @param expression std::string, color expression
    static Fl_Color darker(std::string expression);

    /// @brief Color modification function: multiply color intensity
    /// @param expression std::string, color expression
    static Fl_Color shade(std::string expression);


    /// @brief Color modification function: mix two colors
    /// @param expression std::string, color expression
    static Fl_Color mix(std::string expression);

public:
    /// @brief Default constructor
    CThemeColorCollection();

    /// @brief Loads colors from SPTK theme
    /// @param sptkTheme CXmlDOc&, SPTK theme
    void loadFromSptkTheme(CXmlDoc& sptkTheme);

    /// @brief Loads colors from GTK theme
    /// @param gtkTheme CXmlDOc&, GTK theme
    void loadFromGtkTheme(CXmlDoc& gtkTheme);

    /// @brief Returns color value by color index and state
    Fl_Color color(CThemeColorIndex colorIndex,CThemeColorState state) const { return m_colors[colorIndex][state]; }

    /// @brief Returns current theme foreground color
    Fl_Color fgColor(CThemeColorState state) const { return m_colors[THM_FOREGROUND_COLOR][state]; }

    /// @brief Returns current theme background color
    Fl_Color bgColor(CThemeColorState state) const { return m_colors[THM_BACKGROUND_COLOR][state]; }

    /// @brief Returns current theme base color
    Fl_Color baseColor(CThemeColorState state) const { return m_colors[THM_BASE_COLOR][state]; }

    /// @brief Returns current theme text color
    Fl_Color textColor(CThemeColorState state) const { return m_colors[THM_TEXT_COLOR][state]; }
};

/// @}
};

#endif
