/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CThemeColorScrollBar.h  -  description
                             -------------------
    begin                : Thu Jul 10 2008
    copyright            : (C) 2003-2008 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.
 * Neither the name of the <ORGANIZATION> nor the names of its contributors 
   may be used to endorse or promote products derived from this software 
   without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#ifndef __CTHEMESCROLLBAR_H__
#define __CTHEMESCROLLBAR_H__

#include <FL/Fl.h>
#include <FL/Enumerations.h>
#include <sptk3/cxml>
#include <sptk3/gui/CThemeImageCollection.h>
#include <map>
#include <string>

namespace sptk {

/// @addtogroup gui GUI Classes
/// @{

/// @brief Theme scroll bar type - defines the scroll bar direction
enum CThemeScrollBarType { THM_SCROLLBAR_VERTICAL, THM_SCROLLBAR_HORIZONTAL };

/// @brief The complete set of four images to draw a scroll bar
class CScrollBarImages {
public:

    /// @brief Type of the scroll bar image
    enum type {
        TROUGH,  ///< Image of the scroll bar without slider and buttons
        SLIDER   ///< Image of the scroll bar slider
    };

    CThemeScrollBarType     m_orientation;  ///< Scrollbar orientation
    CThemeImageCollection   m_parts[2];     ///< Scrollbar parts
    CThemeImageCollection   m_steppers[2];  ///< Stepper buttons, 0/1 - (left or upper)normal/pressed, 2/3 - (right or buttom)normal/pressed
public:
    /// @brief Default constructor
    CScrollBarImages() {}

    /// @brief Loads the images from SPTK theme files
    /// @param atype CThemeScrollBarType, scroll bar type
    /// @param desaturateInactiveButtons bool, if true then inactive buttons are desaturated
    /// @param externalFile bool, if true then the file belongs to GTK theme and not to SPTK theme tar
    void load(CThemeScrollBarType atype,bool desaturateInactiveButtons,bool externalFile=false);

    /// @brief Clears the scrollbar images for the default theme
    void clear();
};

/// @brief Defines a collection of scrollbar images and parameters
class CThemeScrollBar {
public:
    /// @brief Images for painting a scroll bar (horizontal and vertical)
    ///
    /// Images should have the compatible size.
    /// If you put the slider image on top its frame - they should show
    /// a picture of 100% progress
    static CScrollBarImages m_scrollBar[2];

    static int m_scrollBarTroughBorder; ///< Trough border width
    static int m_scrollBarSliderWidth;  ///< Slider width
    static int m_scrollBarStepperSize;  ///< Button width

    /// @brief Loads GTK theme scrollbar button images from style node
    /// @param xml CXmlDoc&, GTK theme converted to XML
    /// @param orientation std::string, scrollbar button orientation ("UP","DOWN","LEFT","RIGHT")
    /// @param buttonImages CThemeImageCollection&, output button images
    static void loadGtkScrollbarButtons(CXmlDoc& xml,std::string orientation,CThemeImageCollection& buttonImages);

    /// @brief Loads GTK theme scrollbar troughs
    /// @param xml CXmlDoc&, GTK theme converted to XML
    static void loadGtkScrollbarTroughs(CXmlDoc& xml);

    /// @brief Loads GTK theme scrollbar sliders
    /// @param xml CXmlDoc&, GTK theme converted to XML
    static void loadGtkScrollbarSliders(CXmlDoc& xml);

    /// @brief Loads GTK theme scrollbar button images from GTK theme
    /// @param xml CXmlDoc&, GTK theme converted to XML
    static void loadGtkScrollbars(CXmlDoc& xml);

    /// Computes and returns the minimum size of the scroll bar
    /// @param w int&, scroll bar width, may be altered by scroll bar to meet the scroll bar size requirements
    /// @param h int&, scroll bar height, may be altered by scroll bar to meet the scroll bar size requirements
    /// @returns false, if theme has no pixmaps for the scroll bar, or true otherwise
    static bool sizeScrollBar(int& w,int& h);

    /// @brief Draws a scroll bar image.
    ///
    /// The type of the scroll bar is defined by max of (w,h).
    /// @returns false, if theme has no pixmaps for the scroll bar, or true otherwise
    /// @param x int, x-coordinate of the scroll bar
    /// @param y int, y-coordinate of the scroll bar
    /// @param w int, width of the scroll bar
    /// @param h int, width of the scroll bar
    /// @param minimum float, minimum value of the scroll bar
    /// @param maximum float, maximum value of the scroll bar
    /// @param value float, value of the scroll bar
    /// @param size float, size of the scroll bar
    /// @param activeZone int, 0 - no active zone, 1 - button 1, 2 - button 2, 3 - slider, +8 - pushed
    /// @param active bool, true if scrollbar is active (enabled)
    static bool drawScrollBar(int x,int y,int w,int h,float minimum,float maximum,float value,float size,int activeZone,bool active);
};

/// @}
};

#endif
