/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CMySQLDatabase.cpp  -  description
                             -------------------
    begin                : Mon Aug 18 2008
    copyright            : (C) 2008 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
  * Neither the name of the <ORGANIZATION> nor the names of its contributors
    may be used to endorse or promote products derived from this software
    without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#include <sptk3/db/CMySQLDatabase.h>
#include <sptk3/db/CDatabaseField.h>
#include <sptk3/db/CQuery.h>
#include <sptk3/db/CParams.h>

#include <string>
#include <stdio.h>
#include <arpa/inet.h>

#include <iostream>

#if HAVE_MYSQL == 1

using namespace std;
using namespace sptk;

#define MYSQL_THROW(msg) throwException( msg + string(mysql_error(m_connect)) )

namespace sptk {

    /// @brief MysSQL-specific database field
    class CMySQLField : public CDatabaseField {
        friend class CMySQLDatabase;
    public:
        my_bool         m_null;
        my_bool         m_error;
        unsigned long   m_length;
        MYSQL_TIME      m_time;
    public:
        /// @brief Constructor
        /// @param std::string fieldName, field name
        /// @param fieldColumn int, field column number
        /// @param fieldType int, native field type
        /// @param dataType CVariantType, SPTK field data type
        /// @param fieldSize int, maximum field size in bytes
        CMySQLField(std::string fieldName, int fieldColumn, int fieldType, CVariantType dataType, int fieldSize, MYSQL_BIND& bind)
        : CDatabaseField(fieldName, fieldColumn, fieldType, dataType, fieldSize, 4) {
            m_null = 0;
            m_error = 0;
            m_length = 0;
            bind.buffer_type = (enum_field_types) fieldType;
            bind.buffer_length = fieldSize;

            switch (fieldType) {
                case MYSQL_TYPE_BIT:
                case MYSQL_TYPE_TINY:
                case MYSQL_TYPE_SHORT:
                case MYSQL_TYPE_YEAR:
                case MYSQL_TYPE_LONG:
                case MYSQL_TYPE_LONGLONG:
                case MYSQL_TYPE_FLOAT:
                case MYSQL_TYPE_NEWDECIMAL:
                case MYSQL_TYPE_DOUBLE:
                    /// Using fixed size buffer
                    bind.buffer = (void*) &getInt64();
                    bind.buffer_length = 0;
                    break;
                case MYSQL_TYPE_NEWDATE:
                case MYSQL_TYPE_DATE:
                case MYSQL_TYPE_DATETIME:
                case MYSQL_TYPE_TIME:
                case MYSQL_TYPE_TIMESTAMP:
                    /// Using a special MYSQL_TIME buffer
                    bind.buffer = (void*) &m_time;
                    bind.buffer_length = sizeof(MYSQL_TIME);
                    break;
                default:
                    bind.buffer = (void*) getBuffer();
                    break;
            }

            bind.length = &m_length;
            bind.is_null = &m_null;
            bind.error = &m_error;
        }
    };

    class CMySQLParamValues {
        friend class CMySQLStatement;
    protected:
        unsigned       m_size;
        unsigned       m_count;
        CParamVector   m_params;
        MYSQL_BIND*    m_binds;
        unsigned long* m_lengths;

        static const my_bool is_null;
    public:
        CMySQLParamValues() {
            m_count = 0;
            m_size = 0;
            m_binds   = NULL;
            m_lengths = NULL;
            resize(16);
        }

        ~CMySQLParamValues() {
            if (m_size) {
                free(m_binds);
                free(m_lengths);
            }
        }

        void reset() {
            m_count = 0;
        }

        void resize(unsigned sz) {
            if (sz > m_size) {
                m_size = sz;
                m_binds   = (MYSQL_BIND*)    realloc(m_binds,m_size * sizeof(MYSQL_BIND));
                memset(m_binds,0,m_size * sizeof(MYSQL_BIND));
                m_lengths = (unsigned long*) realloc(m_lengths,m_size * sizeof(unsigned long));
                for (unsigned i = 0; i < sz; i++)
                    m_binds[i].length = &m_lengths[i];
            }
        }

        void setParameters(CParamList& params);

        void setParameterValue(uint32_t number, void* value, unsigned long sz) {
            m_lengths[number] = sz;
            m_binds[number].buffer = value;
            if (value)
                m_binds[number].is_null = 0;
            else
                m_binds[number].is_null = (my_bool*)&CMySQLParamValues::is_null;
        }

        void bindParameterBuffer(uint32_t number,CParam& param) {
            MYSQL_BIND& bind = m_binds[number];

            CVariantType paramType = param.dataType();
            CMySQLDatabase::CTypeToMySQLType(paramType, bind.buffer_type);

            m_lengths[number] = param.dataSize();
            switch (paramType) {
                case VAR_BOOL:
                case VAR_INT:
                case VAR_INT64:
                case VAR_FLOAT:
                    bind.buffer = (void*) &param.getInt64();
                    break;

                case VAR_DATE:
                case VAR_DATE_TIME:
                    bind.buffer = (void*) param.conversionBuffer();
                    m_lengths[number] = sizeof(MYSQL_TIME);
                    break;

                case VAR_STRING:
                case VAR_TEXT:
                case VAR_BUFFER:
                    bind.buffer = (void*) param.getBuffer();
                    break;

                default:
                	throw CException("Unsupported parameter type");
            }
            if (param.isNull())
                bind.is_null = (my_bool*)&CMySQLParamValues::is_null;
            else
                bind.is_null = 0;
        }

        unsigned size() const               { return m_count;   }
        const MYSQL_BIND* binds() const     { return m_binds;   }
        const CParamVector& params() const  { return m_params;  }
    };

    class CMySQLStatement {
        MYSQL_STMT*       m_stmt;
        static unsigned   index;
        unsigned          m_cols;
        MYSQL_BIND*       m_fieldBinds;
        bool              m_eof;
        bool              m_fetchTruncated;
    public:
        CMySQLParamValues      m_paramValues;
    public:
        CMySQLStatement(MYSQL* db) {
            m_stmt = mysql_stmt_init(db);
            m_fieldBinds = 0;
            m_eof = true;
        }

        ~CMySQLStatement() {
            if (m_stmt)
                mysql_stmt_close(m_stmt);
        }

        void clear() {
           m_cols = 0;
        }

        void setColumnCount(unsigned cols) {
            if (m_cols != cols) {
                m_cols = cols;
                if (m_fieldBinds)
                    delete [] m_fieldBinds;
                m_fieldBinds = new MYSQL_BIND[cols];
                memset(m_fieldBinds,0,sizeof(MYSQL_BIND)*cols);
            }
        }

        MYSQL_STMT* stmt() const            { return m_stmt; }
        void        fetch();
        bool        eof() const             { return m_eof; }
        bool        fetchTruncated() const  { return m_fetchTruncated; }
        unsigned    colCount() const        { return m_cols; }

        const CParamVector& params() const  { return m_paramValues.m_params;  }

        MYSQL_BIND* fieldBinds() const      { return m_fieldBinds; }
        void doBindFields();
    };

    unsigned CMySQLStatement::index;
}

const my_bool CMySQLParamValues::is_null(1);

//-----------------------------------------------------------------------------------------------
void CMySQLParamValues::setParameters(CParamList& params) {
    params.enumerate(m_params);
    m_count = m_params.size();
    resize(m_count);
}
//-----------------------------------------------------------------------------------------------
void CMySQLStatement::fetch() {
    switch (mysql_stmt_fetch(m_stmt)) {
        case 1: // Error
                throwException(mysql_stmt_error(m_stmt));
                break;
        case MYSQL_NO_DATA: // No mor data
                m_eof = true;
                break;
        case MYSQL_DATA_TRUNCATED:
                m_fetchTruncated = true;
                m_eof = false;
                break;
        default: // Successful reading (including data truncation)
                m_fetchTruncated = false;
                m_eof = false;
                break;
    }
}

void CMySQLStatement::doBindFields() {
    if (mysql_stmt_bind_result(m_stmt,m_fieldBinds))
        throwException(mysql_stmt_error(m_stmt));
//    if (mysql_stmt_store_result(m_stmt))
//        throwException(mysql_stmt_error(m_stmt));
}
//-----------------------------------------------------------------------------------------------
CMySQLDatabase::CMySQLDatabase(string connectionString)
: CDatabase(connectionString) {
    m_connect = 0;
}

CMySQLDatabase::~CMySQLDatabase() {
    try {
        if (m_inTransaction && active())
            rollbackTransaction();
        close();
        while (m_queryList.size()) {
            try {
                CQuery *query = (CQuery *)m_queryList[0];
                query->disconnect();
            } catch (...) {}}
        m_queryList.clear();
    } catch (...) {}
}

void CMySQLDatabase::parseConnectionString(string connectionString,
    string& host, string& user, string& password, string& dbName, uint32_t& port) const
{
    static const CStrings paramNames("host,user,pass,password,name,dbname,database,port",",");
    host = user = password = dbName = "";
    port = 0;
    CStrings pairStrings(connectionString," ");
    for (unsigned i = 0; i < pairStrings.size(); i++) {
        CStrings pair(pairStrings[i],"=");
        if (pair.size() == 2) {
            string value = replaceAll(replaceAll(pair[1],"\"",""),"\'","");
            switch (paramNames.indexOf(lowerCase(pair[0]))) {
            case 0: host = value; break;
            case 1: user = value; break;
            case 2:
            case 3: password = value; break;
            case 4:
            case 5:
            case 6: dbName = value; break;
            case 7: port = string2int(value,0); break;
            }
        }
    }
}

void CMySQLDatabase::openDatabase(const string newConnectionString) throw(CException) {
    if (!active()) {
        m_inTransaction = false;
        if (newConnectionString.length())
            m_connString = newConnectionString;

        if ( ! ( m_connect = mysql_init (m_connect) ) )
            MYSQL_THROW("mysql_init failed: ");

        parseConnectionString(m_connString,m_host, m_user, m_password, m_dbName, m_port);
        if ( !mysql_real_connect (m_connect, m_host.c_str(), m_user.c_str(), m_password.c_str(), m_dbName.c_str(), m_port, NULL, 0 ) ) {
            m_connect = NULL;
            MYSQL_THROW("mysql_real_connect failed: ");
        }
    }
}

void CMySQLDatabase::closeDatabase() throw(CException) {
    for (unsigned i = 0; i < m_queryList.size(); i++) {
        try {
            CQuery *query = (CQuery *)m_queryList[i];
            queryFreeStmt(query);
        } catch (...) {}
    }
    if (m_connect)
        mysql_close (m_connect);
    m_connect = NULL;
}

void* CMySQLDatabase::handle() const {
    return m_connect;
}

bool CMySQLDatabase::active() const {
    return m_connect != 0L;
}

void CMySQLDatabase::driverBeginTransaction() throw(CException) {
    if (!m_connect)
        open();

    if (m_inTransaction)
        throw CException("Transaction already started.");

    if (mysql_query(m_connect,"START TRANSACTION") != 0)
        MYSQL_THROW("START TRANSACTION failed: ");

    MYSQL_RES *res = mysql_store_result(m_connect);
    if (res)
        mysql_free_result(res);

    m_inTransaction = true;
}

void CMySQLDatabase::driverEndTransaction(bool commit) throw(CException) {
    if (!m_inTransaction)
        throw CException("Transaction isn't started.");

    string action;
    if (commit)
        action = "COMMIT";
    else
        action = "ROLLBACK";

    if (mysql_query(m_connect,action.c_str()) != 0)
        MYSQL_THROW(action + " failed: ");

    MYSQL_RES *res = mysql_store_result(m_connect);
    if (res)
        mysql_free_result(res);

    m_inTransaction = false;
}

//-----------------------------------------------------------------------------------------------
static inline bool successful(int ret) {
    return ret == 0;
}

string CMySQLDatabase::queryError(const CQuery *query) const {
    return mysql_error(m_connect);
}

// Doesn't actually allocate stmt, but makes sure
// the previously allocated stmt is released
void CMySQLDatabase::queryAllocStmt(CQuery *query) {
    queryFreeStmt(query);
    querySetStmt(query, new CMySQLStatement(m_connect));
}

void CMySQLDatabase::queryFreeStmt(CQuery *query) {
    CMySQLLock lock(this);

    CMySQLStatement* statement = (CMySQLStatement*)query->statement();
    if (statement) {
        delete statement;
        querySetStmt(query, 0L);
        querySetPrepared(query, false);
    }
}

void CMySQLDatabase::queryCloseStmt(CQuery *query) {
    CMySQLLock lock(this);

    CMySQLStatement* statement = (CMySQLStatement*)query->statement();
    mysql_stmt_free_result(statement->stmt());
}

void CMySQLDatabase::queryPrepare(CQuery *query) {
    queryFreeStmt(query);

    CMySQLLock lock(this);

    querySetStmt(query, new CMySQLStatement(m_connect));

    CMySQLStatement* statement = (CMySQLStatement*)query->statement();

    CMySQLParamValues& params = statement->m_paramValues;
    params.setParameters(query->params());

    //unsigned   paramCount = params.size();

    if (mysql_stmt_prepare(statement->stmt(),query->sql().c_str(),query->sql().length()) != 0)
        query->logAndThrow("CMySQLDatabase::queryPrepare", mysql_stmt_error(statement->stmt()));

    unsigned fieldCount = mysql_stmt_field_count(statement->stmt());

    statement->setColumnCount(fieldCount);

    querySetPrepared(query, true);
}

void CMySQLDatabase::queryUnprepare(CQuery *query) {
    queryFreeStmt(query);
}

void CMySQLDatabase::queryExecute(CQuery *query) {
    CMySQLLock  lock(this);

    CMySQLStatement* statement = (CMySQLStatement*)query->statement();
    if (mysql_stmt_execute(statement->stmt()))
        query->logAndThrow("CMySQLDatabase::queryExecute", mysql_stmt_error(statement->stmt()));
}

int CMySQLDatabase::queryColCount(CQuery *query) {

    CMySQLStatement* statement = (CMySQLStatement*)query->statement();

    return statement->colCount();
}

void CMySQLDatabase::queryBindParameters(CQuery *query) {
    CMySQLLock lock(this);

    CMySQLStatement*      statement = (CMySQLStatement*)query->statement();
    CMySQLParamValues&    paramValues = statement->m_paramValues;
    const CParamVector&   params = paramValues.params();
    uint32_t              paramNumber = 0;

    for (CParamVector::const_iterator ptor = params.begin(); ptor != params.end(); ptor++, paramNumber++) {
        CParam *param = *ptor;
        CVariantType ptype = param->dataType();
        paramValues.bindParameterBuffer(paramNumber,*param);

        switch (ptype) {
            case VAR_BOOL:
            case VAR_INT:
            case VAR_INT64:
            case VAR_FLOAT:
            case VAR_STRING:
            case VAR_TEXT:
            case VAR_BUFFER:
            break;

            case VAR_DATE: {
                MYSQL_TIME* mysqlTime = (MYSQL_TIME*) param->conversionBuffer();
                int16_t y, m, d;
                param->getDateTime().decodeDate(&y,&m,&d);
                memset(mysqlTime,0,sizeof(MYSQL_TIME));
                mysqlTime->day = d;
                mysqlTime->month = m;
                mysqlTime->year = y;
                mysqlTime->time_type = MYSQL_TIMESTAMP_DATE;
            }
            break;

            case VAR_DATE_TIME: {
                MYSQL_TIME* mysqlTime = (MYSQL_TIME*) param->conversionBuffer();
                int16_t y, m, d;
                int16_t hh, mm, ss, ms;
                param->getDateTime().decodeDate(&y,&m,&d);
                param->getDateTime().decodeTime(&hh,&mm,&ss,&ms);
                memset(mysqlTime,0,sizeof(MYSQL_TIME));
                mysqlTime->day = d;
                mysqlTime->month = m;
                mysqlTime->year = y;
                mysqlTime->hour = hh;
                mysqlTime->minute = mm;
                mysqlTime->second = ss;
                mysqlTime->second_part = ms;
                mysqlTime->time_type = MYSQL_TIMESTAMP_DATETIME;
            }
            break;

            default:
                query->logAndThrow("CMySQLDatabase::queryBindParameters","Unsupported type of parameter "+int2string(paramNumber));
        }
    }

    int resultFormat = 1;   // Results are presented in binary format

    if (!statement->colCount())
        resultFormat = 0;   // VOID result or NO results, using text format

    /// Bind the buffers
    if (mysql_stmt_bind_param(statement->stmt(), (MYSQL_BIND*) paramValues.binds()))
        query->logAndThrow("CMySQLDatabase::queryBindParameters",mysql_stmt_error(statement->stmt()));
}

bool CMySQLDatabase::isMySQLTypeBuffer(enum_field_types mysqlType) {
    switch (mysqlType) {
        case MYSQL_TYPE_BIT:
        case MYSQL_TYPE_TINY:
        case MYSQL_TYPE_SHORT:
        case MYSQL_TYPE_YEAR:
        case MYSQL_TYPE_LONG:
        case MYSQL_TYPE_LONGLONG:
        case MYSQL_TYPE_FLOAT:
        case MYSQL_TYPE_NEWDECIMAL:
        case MYSQL_TYPE_DOUBLE:
            return false;
        default:
            return true;
    }
}

void CMySQLDatabase::MySQLTypeToCType(enum_field_types mysqlType, CVariantType& dataType) {
    switch (mysqlType) {
        case MYSQL_TYPE_BIT:
        case MYSQL_TYPE_TINY:
        case MYSQL_TYPE_SHORT:
        case MYSQL_TYPE_YEAR:
        case MYSQL_TYPE_LONG:      dataType = VAR_INT;       return;
        case MYSQL_TYPE_LONGLONG:  dataType = VAR_INT64;     return;
        case MYSQL_TYPE_FLOAT:
        case MYSQL_TYPE_NEWDECIMAL:
        case MYSQL_TYPE_DOUBLE:    dataType = VAR_FLOAT;     return;
        case MYSQL_TYPE_TINY_BLOB:
        case MYSQL_TYPE_MEDIUM_BLOB:
        case MYSQL_TYPE_LONG_BLOB:
        case MYSQL_TYPE_BLOB:      dataType = VAR_BUFFER;    return;
        case MYSQL_TYPE_NEWDATE:
        case MYSQL_TYPE_DATE:      dataType = VAR_DATE;      return;
        case MYSQL_TYPE_DATETIME:
        case MYSQL_TYPE_TIME:
        case MYSQL_TYPE_TIMESTAMP: dataType = VAR_DATE_TIME; return;
        default:                   dataType = VAR_STRING;    return;
    }
}

void CMySQLDatabase::CTypeToMySQLType(CVariantType dataType, enum_field_types& mysqlType) {
    switch (dataType) {
        case VAR_INT:       mysqlType = MYSQL_TYPE_LONG;    return;        /// Integer 4 bytes
        case VAR_FLOAT:
        case VAR_MONEY:     mysqlType = MYSQL_TYPE_FLOAT;   return;        /// Floating-point (double)
        case VAR_STRING:    mysqlType = MYSQL_TYPE_VARCHAR; return;        /// Varchar
        case VAR_TEXT:
        case VAR_BUFFER:    mysqlType = MYSQL_TYPE_BLOB;    return;        /// BLOB
        case VAR_DATE:
        case VAR_DATE_TIME: mysqlType = MYSQL_TYPE_TIMESTAMP; return;      /// Timestamp
        case VAR_INT64:     mysqlType = MYSQL_TYPE_LONGLONG; return;       /// Integer 8 bytes
        case VAR_BOOL:      mysqlType = MYSQL_TYPE_TINY; return;           /// Boolean
        default:            throwException("Unsupported SPTK data type: " + int2string(dataType));
    }
}

void CMySQLDatabase::queryOpen(CQuery *query) {
    if (!active()) open();

    if (query->active()) return;

    if (!query->statement())
        queryAllocStmt(query);

    if (!query->prepared())
        queryPrepare(query);

    // Bind parameters also executes a query
    queryBindParameters(query);

    //query->fields().clear();

    CMySQLStatement* statement = (CMySQLStatement*)query->statement();
    //if (statement->rowCount() == 0)
    //    return;

    short count = queryColCount(query);
    if (count) {
        querySetActive(query, true);
        if (query->fieldCount() == 0) {
           CMySQLLock lock(this);
           // Reading the column attributes
           char  columnName[256];
           //long  columnType;
           //CVariantType dataType;
           MYSQL_STMT* stmt = statement->stmt();
           MYSQL_RES* metadata = mysql_stmt_result_metadata(stmt);
           MYSQL_BIND* fieldBinds = statement->fieldBinds();
           if (!metadata)
               query->logAndThrow("CMySQLDatabase::queryOpen","query returned no metadata");
           for (short column = 0; column < count; column++) {
               MYSQL_FIELD *field = mysql_fetch_field(metadata);
               strncpy(columnName, field->name, 255);
               columnName[255] = 0;
               if (columnName[0] == 0)
                   sprintf(columnName, "column%02i", column);
               enum_field_types dataType = field->type;
               CVariantType fieldType;
               MySQLTypeToCType(dataType, fieldType);
               unsigned fieldLength = field->length <= 256 ? field->length : 256;
               CMySQLField* fld = new CMySQLField(columnName, column, dataType, fieldType, fieldLength,fieldBinds[column]);
               query->fields().push_back(fld);
           }
        }
        statement->doBindFields();
    }

    queryExecute(query);
    if (count) {
        queryFetch(query);
        querySetEof(query, statement->eof());
    }
}

void CMySQLDatabase::queryFetch(CQuery *query) {
    if (!query->active())
        query->logAndThrow("CMySQLDatabase::queryFetch","Dataset isn't open");

    CMySQLLock lock(this);

    CMySQLStatement* statement = (CMySQLStatement*) query->statement();

    statement->fetch();
    if (statement->eof()) {
        querySetEof(query, true);
        return;
    }

    int fieldCount = query->fieldCount();
    unsigned long dataLength = 0;

    if (!fieldCount)
        return;

    CMySQLField*      field = 0;
    MYSQL_STMT*       stmt = statement->stmt();
    MYSQL_BIND*       binds = statement->fieldBinds();
    for (int column = 0; column < fieldCount; column++) {
        try {
            MYSQL_BIND& bind = binds[column];
            field = (CMySQLField*) & (*query)[(int)column];
            short fieldType = (short) field->dataType();

            bool  isNull = false;
            dataLength = *bind.length;
            if (!dataLength) {
                if (fieldType & (VAR_STRING|VAR_TEXT|VAR_BUFFER))
                    isNull = bind.is_null;
                else
                    isNull = true;
                if (isNull)
                    field->setNull();
                else
                    field->setExternalString("", 0);
            } else {
                switch (fieldType) {
                case VAR_DATE: {
                        MYSQL_TIME* mysqlTime = (MYSQL_TIME*) bind.buffer;
                        CDateTime dt(mysqlTime->year,mysqlTime->month,mysqlTime->day);
                        field->setDate(dt);
                    }
                    break;
                case VAR_DATE_TIME: {
                        MYSQL_TIME* mysqlTime = (MYSQL_TIME*) bind.buffer;
                        CDateTime dt(mysqlTime->year,mysqlTime->month,mysqlTime->day,
                                mysqlTime->hour,mysqlTime->minute,mysqlTime->second);
                        field->setDate(dt);
                    }
                    break;
                case VAR_STRING:
                case VAR_TEXT:
                case VAR_BUFFER:
                    if (bind.buffer_length < dataLength) {
                        /// Buffer was smaller than data, refetch
                        field->checkSize(dataLength);
                        bind.buffer = (void*)field->getBuffer();
                        bind.buffer_length = dataLength+1;
                        if (mysql_stmt_fetch_column(stmt,&bind,column,0))
                            throw runtime_error(mysql_error(m_connect));
                        bind.buffer_length = 0;
                    }
                    break;
                }
                field->dataSize(dataLength);
            }
        } catch (exception& e) {
            query->logAndThrow("CMySQLDatabase::queryFetch","Can't read field "+field->fieldName()+": "+string(e.what()));
        }
    }
}

void CMySQLDatabase::objectList(CDbObjectType objectType, CStrings& objects) throw(std::exception) {
    string tablesSQL("SELECT CONCAT(table_schema,'.',table_name) "
            "FROM information_schema.tables "
            "WHERE table_schema NOT IN ('information_schema','mysql') ");
    string objectsSQL;
    objects.clear();
    switch (objectType) {
        case DOT_PROCEDURES:  objectsSQL = "SELECT DISTINCT CONCAT(routine_schema,'.',routine_name) "
                "FROM information_schema.routines "
                "WHERE routine_schema NOT IN ('information_schema','mysql')";
        break;
        case DOT_TABLES:      objectsSQL = tablesSQL + "AND table_type = 'BASE TABLE'"; break;
        case DOT_VIEWS:       objectsSQL = tablesSQL + "AND table_type = 'VIEW'"; break;
        default:              return; // no information about objects of other types
    }
    CQuery query(this, objectsSQL);
    query.open();
    while (!query.eof()) {
        objects.push_back(query[uint32_t(0)].asString());
        query.next();
    }
    query.close();
}

string CMySQLDatabase::driverDescription() const {
    string version("MySQL ");
    if (m_connect)
        version += mysql_get_server_info(m_connect);
    return version;
}

string CMySQLDatabase::paramMark(unsigned paramIndex) {
    return "?";
}

#endif
