/***************************************************************************
                          SIMPLY POWERFUL TOOLKIT (SPTK)
                          CThemeColorCollection.cpp  -  description
                             -------------------
    begin                : Sat Jul 03 2008
    copyright            : (C) 2003-2008 by Alexey Parshin. All rights reserved.
    email                : alexeyp@gmail.com
 ***************************************************************************/

/***************************************************************************
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.
 * Neither the name of the <ORGANIZATION> nor the names of its contributors 
   may be used to endorse or promote products derived from this software 
   without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ***************************************************************************/

#include <FL/Fl.h>
#include <FL/Enumerations.H>

#include <sptk3/gui/CThemeColorCollection.h>
#include <iostream>

using namespace std;
using namespace sptk;

std::map<std::string,gtk_color_function>     CThemeColorCollection::m_gtkColorFunctionMap;
std::map<std::string,Fl_Color>               CThemeColorCollection::m_colorMap;

CThemeColorCollection::CThemeColorCollection() {
    if (m_gtkColorFunctionMap.empty()) {
        m_gtkColorFunctionMap["darker"] = darker;
        m_gtkColorFunctionMap["lighter"] = lighter;
        m_gtkColorFunctionMap["shade"] = shade;
        m_gtkColorFunctionMap["mix"] = mix;
    }
}

static Fl_Color colorFromHexString(string colorStr) {
    char* eptr;

    uint32_t rgbColor = strtol(colorStr.c_str(),&eptr,16);
    uint32_t blue, green, red;

    if (colorStr.length() == 3) {
        blue  = rgbColor % 16 * 16; rgbColor /= 16;
        green = rgbColor % 16 * 16; rgbColor /= 16;
        red   = rgbColor % 16 * 16;
        rgbColor = (red * 256 + green) * 256 + blue;
    }

    return (Fl_Color) (rgbColor << 8);
}

static void splitArguments(const string& expression,CStrings& arguments) {
    bool  completed = false;
    int   subExpressionLevel = 0;

    const char* ptr1 = expression.c_str();
    const char* ptr2;
    while (ptr2 = strpbrk(ptr1,",(")) {
        if (*ptr2 == '(') {
            const char* ptr = ptr2 + 1;
            subExpressionLevel++;
            while (subExpressionLevel) {
                ptr2 = strpbrk(ptr,"()");
                if (*ptr2 == '(')
                    subExpressionLevel++;
                else
                    subExpressionLevel--;
                ptr = ptr2 + 1;
            }
            ptr2 = strpbrk(ptr,",");
            if (!ptr2)
                break;
        }
        string part(ptr1,ptr2-ptr1);
        part = trim(part);
        if (part[0] == '\"')
            part = part.substr(1,part.length()-2);
        arguments.push_back(part);
        ptr1 = ptr2 + 1;
    }
    arguments.push_back(ptr1);
}

static void splitExpression(const string& colorValue,string& function,string& arguments) {
    size_t pos1 = colorValue.find("(");
    size_t pos2 = colorValue.rfind(")");
    if (pos1 != string::npos && pos2 != string::npos ) {
        function = trim(colorValue.substr(0,pos1-1));
        arguments = trim(colorValue.substr(pos1+1,pos2-pos1-1));
        if (arguments[0] == '\"')
            arguments = arguments.substr(1,arguments.length() - 2);
    }
}

Fl_Color CThemeColorCollection::passby(std::string expression) {
    return gtkColorFunction(expression);
}

Fl_Color CThemeColorCollection::lighter(std::string expression) {
    return fl_lighter(gtkColorFunction(expression));
}

Fl_Color CThemeColorCollection::darker(std::string expression) {
    return fl_darker(gtkColorFunction(expression));
}

Fl_Color CThemeColorCollection::mix(std::string expression) {
    CStrings args;
    splitArguments(expression,args);
    double weight = atof(args[0].c_str());
    Fl_Color color1 = gtkColorFunction(args[1]);
    Fl_Color color2 = gtkColorFunction(args[2]);
    return fl_color_average(color1,color2,weight);
}

Fl_Color CThemeColorCollection::shade(std::string expression) {
    CStrings args;
    splitArguments(expression,args);
    if (args.size() < 2)
        return fl_darker(gtkColorFunction(expression));
    double multiplier = atof(args[0].c_str());
    if (multiplier < 0.001)
        multiplier = 1;

    Fl_Color color = gtkColorFunction(args[1]);
    if (color > 1)
        return fl_color_average(FL_WHITE,color,multiplier-1);
    return fl_color_average(color,FL_BLACK,multiplier);
}

Fl_Color CThemeColorCollection::gtkColorFunction(std::string expression) {
    string function, arguments;
    string colorValue(trim(expression));
    if (colorValue[0] == '\"')
        colorValue = colorValue.substr(1,colorValue.length()-2);
    switch (colorValue[0]) {
        case '#': // Direct color value
            return colorFromHexString(colorValue.c_str()+1);
        case '@': // Color from color map
            return m_colorMap[colorValue.c_str()+1];
        default:  // Color function
            splitExpression(colorValue,function,arguments);
            if (function.empty()) {
                cerr << "Color expression '" << expression << "' isn't understood" << endl;
                return FL_BLACK;
            }
            std::map<std::string,gtk_color_function>::iterator itor = m_gtkColorFunctionMap.find(function);
            if (itor == m_gtkColorFunctionMap.end()) {
                cerr << "Color function '" << function << "' isn't supported" << endl;
                return passby(arguments);
            } else {
                gtk_color_function cfunction = itor->second;
                return cfunction(arguments);
            }
            break;
    }
}

void CThemeColorCollection::loadColor(CXmlNode* colorNode,CThemeColorIndex colorIndex) {
    static const CStrings colorStateNames("NORMAL,PRELIGHT,SELECTED,ACTIVE,INSENSITIVE",",");
    CXmlAttributes::iterator itor = colorNode->attributes().begin();
    for (; itor != colorNode->attributes().end(); itor++) {
        CXmlNode* colorStateNode = *itor;
        CThemeColorState colorState = (CThemeColorState) colorStateNames.indexOf(colorStateNode->name());
        if (colorState == THM_COLOR_UNDEFINED)
            continue;
        string colorValue(colorStateNode->value());
        Fl_Color color = gtkColorFunction(colorValue);
        m_colors[colorIndex][colorState] = color;
    }
}

static const char* colorNames[THM_MAX_COLOR_INDEX] = { "fg", "bg", "base", "text" };

void CThemeColorCollection::loadFromSptkTheme(CXmlDoc& sptkTheme) {
    loadColorMap(sptkTheme,"/color_scheme");
    for (unsigned colorIndex = 0; colorIndex < THM_MAX_COLOR_INDEX; colorIndex++) {
        string colorXPath = string("/color_scheme/") + colorNames[colorIndex];
        CXmlNodeVector colorNodes;
        sptkTheme.select(colorNodes,colorXPath);
        if (colorNodes.size() == 1) {
            CXmlNode* colorNode = *(colorNodes.begin());
            loadColor(colorNode,CThemeColorIndex(colorIndex));
        }
    }
}

void CThemeColorCollection::loadFromGtkTheme(CXmlDoc& gtkTheme) {
    if (!loadColorMap(gtkTheme,"/gtk_color_scheme"))
        loadColorMap(gtkTheme,"/gtk-color-scheme");

    string stylesXPath = "/styles/style";
    CXmlNodeVector styleNodes;
    gtkTheme.select(styleNodes,stylesXPath);
    CXmlNode* defaultStyleNode = *styleNodes.begin();
    for (CXmlNode::iterator itor = styleNodes.begin(); itor != styleNodes.end(); itor++) {
        CXmlNode* styleNode = *itor;
        string styleName(styleNode->getAttribute("name"));
        if (styleName == "default" || styleName.find("-default") != string::npos) {
            defaultStyleNode = styleNode;
            break;
        }
    }

    for (unsigned colorIndex = 0; colorIndex < THM_MAX_COLOR_INDEX; colorIndex++) {
        string colorXPath(colorNames[colorIndex]);
        CXmlNodeVector colorNodes;
        defaultStyleNode->select(colorNodes,colorXPath);
        unsigned elements = colorNodes.size();
        if (elements == 1) {
            CXmlNode* colorNode = *(colorNodes.begin());
            loadColor(colorNode,CThemeColorIndex(colorIndex));
        }
    }

    Fl::set_color(FL_FOREGROUND_COLOR,textColor(THM_COLOR_NORMAL));
    Fl::set_color(FL_BACKGROUND_COLOR,bgColor(THM_COLOR_NORMAL));
    Fl::set_color(FL_BACKGROUND2_COLOR,baseColor(THM_COLOR_NORMAL));
    Fl::set_color(FL_INACTIVE_COLOR,fgColor(THM_COLOR_INSENSITIVE));
    Fl::set_color(FL_SELECTION_COLOR,bgColor(THM_COLOR_SELECTED));

    Fl_Color backgroundColor = bgColor(THM_COLOR_NORMAL);
    Fl_Color darkColor = fl_darker(backgroundColor);
    Fl::set_color(FL_DARK1,darkColor); darkColor = fl_darker(darkColor);
    Fl::set_color(FL_DARK2,darkColor); darkColor = fl_darker(darkColor);
    Fl::set_color(FL_DARK3,darkColor);
    Fl_Color lightColor = fl_lighter(backgroundColor);
    Fl::set_color(FL_LIGHT1,lightColor); lightColor = fl_lighter(lightColor);
    Fl::set_color(FL_LIGHT2,lightColor); lightColor = fl_lighter(lightColor);
    Fl::set_color(FL_LIGHT3,lightColor);
}

bool CThemeColorCollection::loadColorMap(CXmlDoc& gtkTheme,string colorMapXPath) {
    m_colorMap.clear();

    CXmlNodeVector colorMapNodes;
    gtkTheme.select(colorMapNodes,colorMapXPath);
    if (!colorMapNodes.size())
        return false;

    CXmlNode* colorMapNode = *(colorMapNodes.begin());

    string colorValues = colorMapNode->getAttribute("value");
    if (colorValues.empty())
        colorValues = colorMapNode->getAttribute("colors");
    CStrings colorMapStrings(colorValues,"\\n");

    for (unsigned i = 0; i < colorMapStrings.size(); i++) {
        CStrings colorInfo(colorMapStrings[i],":#");
        if (colorInfo.size() != 2)
            continue;
        m_colorMap[colorInfo[0]] = colorFromHexString(colorInfo[1]);
    }
    return true;
}
